@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,8 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
 
 #include "php.h"
@@ -21,6 +19,13 @@
 #include "php_globals.h"
 #include "zend_closures.h"
 #include "zend_exceptions.h"
+#include "zend_interfaces.h"
+
+#if PHP_VERSION_ID >= 80100
+# include "base_private.h"
+# include "Zend/zend_fibers.h"
+# include "Zend/zend_observer.h"
+#endif
 
 #include "php_xdebug.h"
 #include "php_xdebug_arginfo.h"
@@ -38,7 +43,11 @@
 
 ZEND_EXTERN_MODULE_GLOBALS(xdebug)
 
-#if PHP_VERSION_ID >= 80000
+#if PHP_VERSION_ID >= 80100
+void (*xdebug_old_error_cb)(int type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message);
+void (*xdebug_new_error_cb)(int type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message);
+static void xdebug_error_cb(int orig_type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message);
+#elif PHP_VERSION_ID >= 80000
 void (*xdebug_old_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message);
 void (*xdebug_new_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message);
 static void xdebug_error_cb(int orig_type, const char *error_filename, const uint32_t error_lineno, zend_string *message);
@@ -84,13 +93,6 @@ static zend_op_array *xdebug_compile_file(zend_file_handle *file_handle, int typ
 }
 /* }}} */
 
-static void xdebug_declared_var_dtor(void *dummy, void *elem)
-{
-	xdebug_str *s = (xdebug_str*) elem;
-
-	xdebug_str_free(s);
-}
-
 /* I don't like this API, but the function_stack_entry does not keep this as a
  * pointer, and hence we need two APIs for freeing :-S */
 void xdebug_func_dtor_by_ref(xdebug_func *elem)
@@ -98,8 +100,11 @@ void xdebug_func_dtor_by_ref(xdebug_func *elem)
 	if (elem->function) {
 		xdfree(elem->function);
 	}
-	if (elem->class_name) {
-		zend_string_release(elem->class_name);
+	if (elem->object_class) {
+		zend_string_release(elem->object_class);
+	}
+	if (elem->scope_class) {
+		zend_string_release(elem->scope_class);
 	}
 }
 
@@ -145,38 +150,6 @@ static void function_stack_entry_dtor(void *elem)
 	}
 }
 
-static void add_used_variables(function_stack_entry *fse, zend_op_array *op_array)
-{
-	unsigned int i = 0;
-
-	if (!fse->declared_vars) {
-		fse->declared_vars = xdebug_llist_alloc(xdebug_declared_var_dtor);
-	}
-
-	/* gather used variables from compiled vars information */
-	while (i < (unsigned int) op_array->last_var) {
-		xdebug_llist_insert_next(fse->declared_vars, XDEBUG_LLIST_TAIL(fse->declared_vars), xdebug_str_create(STR_NAME_VAL(op_array->vars[i]), STR_NAME_LEN(op_array->vars[i])));
-		i++;
-	}
-
-	/* opcode scanning time */
-	while (i < op_array->last) {
-		char *cv = NULL;
-		int cv_len;
-
-		if (op_array->opcodes[i].op1_type == IS_CV) {
-			cv = (char *) xdebug_get_compiled_variable_name(op_array, op_array->opcodes[i].op1.var, &cv_len);
-			xdebug_llist_insert_next(fse->declared_vars, XDEBUG_LLIST_TAIL(fse->declared_vars), xdebug_str_create(cv, cv_len));
-		}
-		if (op_array->opcodes[i].op2_type == IS_CV) {
-			cv = (char *) xdebug_get_compiled_variable_name(op_array, op_array->opcodes[i].op2.var, &cv_len);
-			xdebug_llist_insert_next(fse->declared_vars, XDEBUG_LLIST_TAIL(fse->declared_vars), xdebug_str_create(cv, cv_len));
-		}
-		i++;
-	}
-}
-
-
 int xdebug_include_or_eval_handler(XDEBUG_OPCODE_HANDLER_ARGS)
 {
 	const zend_op *opline = execute_data->opline;
@@ -242,22 +215,25 @@ void xdebug_build_fname(xdebug_func *tmp, zend_execute_data *edata)
 		tmp->type = XFUNC_NORMAL;
 		if ((Z_TYPE(edata->This)) == IS_OBJECT) {
 			tmp->type = XFUNC_MEMBER;
-			if (edata->func->common.scope && strcmp(edata->func->common.scope->name->val, "class@anonymous") == 0) {
-				char *tmp_class_name = xdebug_sprintf(
+			if (edata->func->common.scope && strstr(edata->func->common.scope->name->val, "@anonymous") != NULL) {
+				char *tmp_object_class = xdebug_sprintf(
 					"{anonymous-class:%s:%d-%d}",
 					edata->func->common.scope->info.user.filename->val,
 					edata->func->common.scope->info.user.line_start,
 					edata->func->common.scope->info.user.line_end
 				);
-				tmp->class_name = zend_string_init(tmp_class_name, strlen(tmp_class_name), 0);
-				xdfree(tmp_class_name);
+				tmp->object_class = zend_string_init(tmp_object_class, strlen(tmp_object_class), 0);
+				xdfree(tmp_object_class);
 			} else {
-				tmp->class_name = zend_string_copy(edata->This.value.obj->ce->name);
+				if (edata->func->common.scope) {
+					tmp->scope_class = zend_string_copy(edata->func->common.scope->name);
+				}
+				tmp->object_class = zend_string_copy(edata->This.value.obj->ce->name);
 			}
 		} else {
 			if (edata->func->common.scope) {
 				tmp->type = XFUNC_STATIC_MEMBER;
-				tmp->class_name = zend_string_copy(edata->func->common.scope->name);
+				tmp->object_class = zend_string_copy(edata->func->common.scope->name);
 			}
 		}
 		if (edata->func->common.function_name) {
@@ -313,6 +289,7 @@ normal_after_all:
 		} else if (
 			edata &&
 			edata->prev_execute_data &&
+			edata->prev_execute_data->func &&
 			edata->prev_execute_data->func->type == ZEND_USER_FUNCTION &&
 			edata->prev_execute_data->opline &&
 			edata->prev_execute_data->opline->opcode == ZEND_INCLUDE_OR_EVAL
@@ -589,7 +566,7 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 	zdata = EG(current_execute_data);
 
 	tmp = (function_stack_entry*) xdebug_vector_push(XG_BASE(stack));
-	tmp->level         = XG_BASE(level);
+	tmp->level         = XDEBUG_VECTOR_COUNT(XG_BASE(stack));
 	tmp->user_defined  = type;
 	tmp->op_array      = op_array;
 
@@ -618,7 +595,7 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 	}
 
 	if (!tmp->filename) {
-		tmp->filename = zend_string_init("UNKNOWN?", sizeof("UNKNOWN?") - 1, 0);
+		tmp->filename = zend_string_init("Unknown", sizeof("Unknown") - 1, 0);
 	}
 	tmp->lineno = 0;
 
@@ -637,9 +614,10 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 
 	xdebug_build_fname(&(tmp->function), zdata);
 	if (!tmp->function.type) {
-		tmp->function.function   = xdstrdup("{main}");
-		tmp->function.class_name = NULL;
-		tmp->function.type       = XFUNC_MAIN;
+		tmp->function.function     = xdstrdup("{main}");
+		tmp->function.object_class = NULL;
+		tmp->function.scope_class  = NULL;
+		tmp->function.type         = XFUNC_MAIN;
 
 	} else if (tmp->function.type & XFUNC_INCLUDES) {
 		tmp->lineno = 0;
@@ -668,8 +646,8 @@ function_stack_entry *xdebug_add_stack_frame(zend_execute_data *zdata, zend_op_a
 		}
 	}
 
-	/* Now we have location and name, we can run the filter */
-	xdebug_filter_run_tracing(tmp);
+	/* Now we have location and name, we can run the filter (for stack and tracing)*/
+	xdebug_filter_run(tmp);
 
 	/* Count code coverage line for call */
 	xdebug_coverage_count_line_if_branch_check_active(op_array, tmp->filename, tmp->lineno);
@@ -717,7 +695,7 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 		return;
 	}
 
-	if (XG_BASE(in_execution) && XG_BASE(level) == 0) {
+	if (XG_BASE(in_execution) && XDEBUG_VECTOR_COUNT(XG_BASE(stack)) == 0) {
 		if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
 			xdebug_debugger_set_program_name(op_array->filename);
 			xdebug_debug_init_if_requested_at_startup();
@@ -736,8 +714,7 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 		}
 	}
 
-	XG_BASE(level)++;
-	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP) && (signed long) XG_BASE(level) > XINI_BASE(max_nesting_level) && (XINI_BASE(max_nesting_level) != -1)) {
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP) && (signed long) XDEBUG_VECTOR_COUNT(XG_BASE(stack)) >= XINI_BASE(max_nesting_level) && (XINI_BASE(max_nesting_level) != -1)) {
 		zend_throw_exception_ex(zend_ce_error, 0, "Xdebug has detected a possible infinite loop, and aborted your script with a stack depth of '" ZEND_LONG_FMT "' frames", XINI_BASE(max_nesting_level));
 	}
 
@@ -768,13 +745,13 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 		 * variables in include/required files to all the stack levels above, until
 		 * we hit a function or the top level stack.  This is so that the variables
 		 * show up correctly where they should be.  We always call
-		 * add_used_variables on the current stack level, otherwise vars in include
-		 * files do not show up in the locals list.  */
+		 * xdebug_lib_register_compiled_variables on the current stack level,
+		 * otherwise vars in include files do not show up in the locals list. */
 		function_stack_entry *loop_fse = XDEBUG_VECTOR_TAIL(XG_BASE(stack));
 		int                   i;
 
 		for (i = 0; i < XDEBUG_VECTOR_COUNT(XG_BASE(stack)); i++, loop_fse--) {
-			add_used_variables(loop_fse, op_array);
+			xdebug_lib_register_compiled_variables(loop_fse, op_array);
 			if (XDEBUG_IS_NORMAL_FUNCTION(&loop_fse->function)) {
 				break;
 			}
@@ -828,13 +805,12 @@ static void xdebug_execute_ex(zend_execute_data *execute_data)
 	if (XG_BASE(stack)) {
 		xdebug_vector_pop(XG_BASE(stack));
 	}
-	XG_BASE(level)--;
 }
 
 static int check_soap_call(function_stack_entry *fse, zend_execute_data *execute_data)
 {
 	if (
-		fse->function.class_name &&
+		fse->function.object_class &&
 		Z_OBJ(EX(This)) &&
 		Z_TYPE(EX(This)) == IS_OBJECT &&
 		(zend_hash_str_find_ptr(&module_registry, "soap", sizeof("soap") - 1) != NULL)
@@ -865,7 +841,9 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 	int                   function_nr = 0;
 	int                   function_call_traced = 0;
 	int                   restore_error_handler_situation = 0;
-#if PHP_VERSION_ID >= 80000
+#if PHP_VERSION_ID >= 80100
+	void                (*tmp_error_cb)(int type, zend_string *error_filename, const uint32_t error_lineno, zend_string *message) = NULL;
+#elif PHP_VERSION_ID >= 80000
 	void                (*tmp_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, zend_string *message) = NULL;
 #else
 	void                (*tmp_error_cb)(int type, const char *error_filename, const uint32_t error_lineno, const char *format, va_list args) ZEND_ATTRIBUTE_PTR_FORMAT(printf, 4, 0) = NULL;
@@ -881,8 +859,7 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 		return;
 	}
 
-	XG_BASE(level)++;
-	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP) && (signed long) XG_BASE(level) > XINI_BASE(max_nesting_level) && (XINI_BASE(max_nesting_level) != -1)) {
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP) && (signed long) XDEBUG_VECTOR_COUNT(XG_BASE(stack)) >= XINI_BASE(max_nesting_level) && (XINI_BASE(max_nesting_level) != -1)) {
 		zend_throw_exception_ex(zend_ce_error, 0, "Xdebug has detected a possible infinite loop, and aborted your script with a stack depth of '" ZEND_LONG_FMT "' frames", XINI_BASE(max_nesting_level));
 	}
 
@@ -948,7 +925,6 @@ static void xdebug_execute_internal(zend_execute_data *current_execute_data, zva
 	if (XG_BASE(stack)) {
 		xdebug_vector_pop(XG_BASE(stack));
 	}
-	XG_BASE(level)--;
 }
 
 static void xdebug_base_overloaded_functions_setup(void)
@@ -957,14 +933,22 @@ static void xdebug_base_overloaded_functions_setup(void)
 
 	/* Override set_time_limit with our own function to prevent timing out while debugging */
 	orig = zend_hash_str_find_ptr(EG(function_table), "set_time_limit", sizeof("set_time_limit") - 1);
-	XG_BASE(orig_set_time_limit_func) = orig->internal_function.handler;
-	orig->internal_function.handler = zif_xdebug_set_time_limit;
+	if (orig) {
+		XG_BASE(orig_set_time_limit_func) = orig->internal_function.handler;
+		orig->internal_function.handler = zif_xdebug_set_time_limit;
+	} else {
+		XG_BASE(orig_set_time_limit_func) = NULL;
+	}
 
 	/* Override error_reporting with our own function, to be able to give right answer during DBGp's
 	 * 'eval' commands */
 	orig = zend_hash_str_find_ptr(EG(function_table), "error_reporting", sizeof("error_reporting") - 1);
-	XG_BASE(orig_error_reporting_func) = orig->internal_function.handler;
-	orig->internal_function.handler = zif_xdebug_error_reporting;
+	if (orig) {
+		XG_BASE(orig_error_reporting_func) = orig->internal_function.handler;
+		orig->internal_function.handler = zif_xdebug_error_reporting;
+	} else {
+		XG_BASE(orig_error_reporting_func) = NULL;
+	}
 
 	/* Override pcntl_exec with our own function to be able to write profiling summary */
 	orig = zend_hash_str_find_ptr(EG(function_table), "pcntl_exec", sizeof("pcntl_exec") - 1);
@@ -990,11 +974,19 @@ static void xdebug_base_overloaded_functions_restore(void)
 {
 	zend_function *orig;
 
-	orig = zend_hash_str_find_ptr(EG(function_table), "set_time_limit", sizeof("set_time_limit") - 1);
-	orig->internal_function.handler = XG_BASE(orig_set_time_limit_func);
+	if (XG_BASE(orig_set_time_limit_func)) {
+		orig = zend_hash_str_find_ptr(EG(function_table), "set_time_limit", sizeof("set_time_limit") - 1);
+		if (orig) {
+			orig->internal_function.handler = XG_BASE(orig_set_time_limit_func);
+		}
+	}
 
-	orig = zend_hash_str_find_ptr(EG(function_table), "error_reporting", sizeof("error_reporting") - 1);
-	orig->internal_function.handler = XG_BASE(orig_error_reporting_func);
+	if (XG_BASE(orig_error_reporting_func)) {
+		orig = zend_hash_str_find_ptr(EG(function_table), "error_reporting", sizeof("error_reporting") - 1);
+		if (orig) {
+			orig->internal_function.handler = XG_BASE(orig_error_reporting_func);
+		}
+	}
 
 	if (XG_BASE(orig_pcntl_exec_func)) {
 		orig = zend_hash_str_find_ptr(EG(function_table), "pcntl_exec", sizeof("pcntl_exec") - 1);
@@ -1021,6 +1013,97 @@ static int xdebug_closure_serialize_deny_wrapper(zval *object, unsigned char **b
 	return FAILURE;
 }
 
+#if PHP_VERSION_ID >= 80100
+static struct xdebug_fiber_entry* xdebug_fiber_entry_ctor(xdebug_vector *stack)
+{
+	struct xdebug_fiber_entry *tmp = xdmalloc(sizeof(struct xdebug_fiber_entry));
+
+	tmp->stack = stack;
+
+	return tmp;
+}
+
+static void xdebug_fiber_entry_dtor(struct xdebug_fiber_entry *entry)
+{
+	xdebug_vector_destroy(entry->stack);
+	xdfree(entry);
+}
+
+static xdebug_str *create_key_for_fiber(zend_fiber_context *fiber)
+{
+	xdebug_str *tmp = xdebug_str_new();
+
+	xdebug_str_add_fmt(tmp, "{fiber:%0" PRIXPTR "}", ((uintptr_t) fiber));
+
+	return tmp;
+}
+
+static void add_fiber_main(zend_fiber_context *fiber)
+{
+	xdebug_str           *key = create_key_for_fiber(fiber);
+	function_stack_entry *tmp = (function_stack_entry*) xdebug_vector_push(XG_BASE(stack));
+
+	tmp->level = 0;
+	tmp->user_defined = XDEBUG_BUILT_IN;
+	tmp->function.type = XFUNC_FIBER;
+	tmp->function.object_class = NULL;
+	tmp->function.scope_class = NULL;
+	tmp->function.function = xdstrdup(key->d);
+}
+
+static xdebug_vector* create_stack_for_fiber(zend_fiber_context *fiber)
+{
+	xdebug_vector             *tmp_stack = xdebug_vector_alloc(sizeof(function_stack_entry), function_stack_entry_dtor);
+	xdebug_str                *key       = create_key_for_fiber(fiber);
+	struct xdebug_fiber_entry *entry     = xdebug_fiber_entry_ctor(tmp_stack);
+
+	xdebug_hash_add(XG_BASE(fiber_stacks), key->d, key->l, entry);
+
+	xdebug_str_free(key);
+
+	return tmp_stack;
+}
+
+static void remove_stack_for_fiber(zend_fiber_context *fiber)
+{
+	xdebug_str *key = create_key_for_fiber(fiber);
+
+	xdebug_hash_delete(XG_BASE(fiber_stacks), key->d, key->l);
+
+	xdebug_str_free(key);
+}
+
+static xdebug_vector *find_stack_for_fiber(zend_fiber_context *fiber)
+{
+	struct xdebug_fiber_entry *entry = NULL;
+	xdebug_str                *key = create_key_for_fiber(fiber);
+
+	xdebug_hash_find(XG_BASE(fiber_stacks), key->d, key->l, (void*) &entry);
+
+	xdebug_str_free(key);
+
+	return entry->stack;
+}
+
+static void xdebug_fiber_switch_observer(zend_fiber_context *from, zend_fiber_context *to)
+{
+	xdebug_vector *current_stack;
+
+	if (from->status == ZEND_FIBER_STATUS_DEAD) {
+		remove_stack_for_fiber(from);
+	}
+	if (to->status == ZEND_FIBER_STATUS_INIT) {
+		current_stack = create_stack_for_fiber(to);
+	} else {
+		current_stack = find_stack_for_fiber(to);
+	}
+	XG_BASE(stack) = current_stack;
+
+	if (to->status == ZEND_FIBER_STATUS_INIT) {
+		add_fiber_main(to);
+	}
+}
+#endif
 
 void xdebug_base_minit(INIT_FUNC_ARGS)
 {
@@ -1046,7 +1129,9 @@ void xdebug_base_minit(INIT_FUNC_ARGS)
 	XG_BASE(error_reporting_overridden) = 0;
 	XG_BASE(output_is_tty) = OUTPUT_NOT_CHECKED;
 
-	xdebug_nanotime_init();
+#if PHP_VERSION_ID >= 80100
+	zend_observer_fiber_switch_register(xdebug_fiber_switch_observer);
+#endif
 }
 
 void xdebug_base_mshutdown()
@@ -1077,8 +1162,12 @@ void xdebug_base_rinit()
 		xdebug_base_use_xdebug_throw_exception_hook();
 	}
 
+#if PHP_VERSION_ID >= 80100
+	XG_BASE(fiber_stacks) = xdebug_hash_alloc(64, (xdebug_hash_dtor_t) xdebug_fiber_entry_dtor);
+	XG_BASE(stack) = create_stack_for_fiber(EG(main_fiber_context));
+#else
 	XG_BASE(stack) = xdebug_vector_alloc(sizeof(function_stack_entry), function_stack_entry_dtor);
-	XG_BASE(level)         = 0;
+#endif
 	XG_BASE(in_debug_info) = 0;
 	XG_BASE(prev_memory)   = 0;
 	XG_BASE(function_count) = -1;
@@ -1099,21 +1188,26 @@ void xdebug_base_rinit()
 	XG_BASE(in_execution) = 1;
 
 	/* filters */
-	XG_BASE(filter_type_tracing)       = XDEBUG_FILTER_NONE;
-	XG_BASE(filter_type_profiler)      = XDEBUG_FILTER_NONE;
 	XG_BASE(filter_type_code_coverage) = XDEBUG_FILTER_NONE;
-	XG_BASE(filters_tracing)           = xdebug_llist_alloc(xdebug_llist_string_dtor);
+	XG_BASE(filter_type_stack)         = XDEBUG_FILTER_NONE;
+	XG_BASE(filter_type_tracing)       = XDEBUG_FILTER_NONE;
 	XG_BASE(filters_code_coverage)     = xdebug_llist_alloc(xdebug_llist_string_dtor);
+	XG_BASE(filters_stack)             = xdebug_llist_alloc(xdebug_llist_string_dtor);
+	XG_BASE(filters_tracing)           = xdebug_llist_alloc(xdebug_llist_string_dtor);
 
 	xdebug_base_overloaded_functions_setup();
 }
 
 void xdebug_base_post_deactivate()
 {
+#if PHP_VERSION_ID >= 80100
+	xdebug_hash_destroy(XG_BASE(fiber_stacks));
+	XG_BASE(fiber_stacks) = NULL;
+#else
 	xdebug_vector_destroy(XG_BASE(stack));
+#endif
 	XG_BASE(stack) = NULL;
 
-	XG_BASE(level)            = 0;
 	XG_BASE(in_debug_info)    = 0;
 
 	if (XG_BASE(last_eval_statement)) {
@@ -1126,8 +1220,9 @@ void xdebug_base_post_deactivate()
 	}
 
 	/* filters */
-	xdebug_llist_destroy(XG_BASE(filters_tracing), NULL);
 	xdebug_llist_destroy(XG_BASE(filters_code_coverage), NULL);
+	xdebug_llist_destroy(XG_BASE(filters_stack), NULL);
+	xdebug_llist_destroy(XG_BASE(filters_tracing), NULL);
 	XG_BASE(filters_tracing) = NULL;
 	XG_BASE(filters_code_coverage) = NULL;
 
@@ -1141,7 +1236,24 @@ void xdebug_base_rshutdown()
 }
 
 /* Error callback for formatting stack traces */
-#if PHP_VERSION_ID >= 80000
+#if PHP_VERSION_ID >= 80100
+static void xdebug_error_cb(int orig_type, zend_string *error_filename, const unsigned int error_lineno, zend_string *message)
+{
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
+		int type                        = orig_type & E_ALL;
+		char *error_type_str            = xdebug_error_type(type);
+
+		xdebug_debugger_error_cb(error_filename, error_lineno, type, error_type_str, ZSTR_VAL(message));
+
+		xdfree(error_type_str);
+	}
+	if (XDEBUG_MODE_IS(XDEBUG_MODE_DEVELOP)) {
+		xdebug_develop_error_cb(orig_type, error_filename, error_lineno, message);
+	} else {
+		xdebug_old_error_cb(orig_type, error_filename, error_lineno, message);
+	}
+}
+#elif PHP_VERSION_ID >= 80000
 static void xdebug_error_cb(int orig_type, const char *error_filename, const unsigned int error_lineno, zend_string *message)
 {
 	if (XDEBUG_MODE_IS(XDEBUG_MODE_STEP_DEBUG)) {
@@ -1237,7 +1349,10 @@ static void xdebug_throw_exception_hook(zval *exception)
 		code_str = xdstrdup("");
 	}
 
-	convert_to_string_ex(message);
+	if (Z_TYPE_P(message) != IS_STRING) {
+		message = NULL;
+	}
+
 	convert_to_string_ex(file);
 	convert_to_long_ex(line);
 
