@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,9 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   |          Thomas Vanhaniemi <thomas.vanhaniemi@arcada.fi>             |
-   +----------------------------------------------------------------------+
  */
 #include "php_xdebug.h"
 #include <string.h>
@@ -353,6 +350,13 @@ static int xdebug_create_socket(const char *hostname, int dport, int timeout)
 #if !WIN32 && !WINNT
 		set_keepalive_options(sockfd);
 #endif
+
+		/* Now we have a socket, update the last seen hostname and port */
+		if (XG_DBG(context).connected_hostname) {
+			xdfree(XG_DBG(context).connected_hostname);
+		}
+		XG_DBG(context).connected_hostname = xdstrdup(hostname);
+		XG_DBG(context).connected_port = dport;
 	}
 
 	return sockfd;
@@ -440,9 +444,9 @@ static void xdebug_init_normal_debugger(xdebug_str *connection_attempts)
 	}
 }
 
-static void xdebug_init_cloud_debugger()
+static void xdebug_init_cloud_debugger(const char *cloud_id)
 {
-	unsigned long  crc = xdebug_crc32(XINI_DBG(cloud_id), strlen(XINI_DBG(cloud_id)));
+	unsigned long  crc = xdebug_crc32(cloud_id, strlen(cloud_id));
 	char          *host;
 
 	host = xdebug_sprintf("%c.cloud.xdebug.com", (crc & 0x0f) + 'a');
@@ -453,6 +457,24 @@ static void xdebug_init_cloud_debugger()
 	xdfree(host);
 }
 
+/**
+ * dXXXXXXa-cXXa-4XX7-9XX3-fXXXXXXXXXX0
+ */
+static int ide_key_is_cloud_id()
+{
+	const char *k = XG_DBG(ide_key);
+
+	if (strlen(k) != 36) {
+		return 0;
+	}
+
+	if (k[8] != '-' || k[13] != '-' || k[18] != '-' || k[23] != '-') {
+		return 0;
+	}
+
+	return 1;
+}
+
 static void xdebug_init_debugger()
 {
 	xdebug_str *connection_attempts = xdebug_str_new();
@@ -460,10 +482,15 @@ static void xdebug_init_debugger()
 	/* Get handler from mode */
 	XG_DBG(context).handler = &xdebug_handler_dbgp;
 
-	if (strcmp(XINI_DBG(cloud_id), "") == 0) {
-		xdebug_init_normal_debugger(connection_attempts);
+	if (strcmp(XINI_DBG(cloud_id), "") != 0) {
+		xdebug_init_cloud_debugger(XINI_DBG(cloud_id));
+		XG_DBG(context).host_type = XDEBUG_CLOUD;
+	} else if (XG_DBG(ide_key) && ide_key_is_cloud_id()) {
+		xdebug_init_cloud_debugger(XG_DBG(ide_key));
+		XG_DBG(context).host_type = XDEBUG_CLOUD_FROM_TRIGGER_VALUE;
 	} else {
-		xdebug_init_cloud_debugger();
+		xdebug_init_normal_debugger(connection_attempts);
+		XG_DBG(context).host_type = XDEBUG_NORMAL;
 	}
 
 	/* Check whether we're connected, or why not */
@@ -583,31 +610,30 @@ static int xdebug_handle_start_session()
 		))
 		&& !SG(headers_sent)
 	) {
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "Found 'XDEBUG_SESSION_START' HTTP variable, with value '%s'", Z_STRVAL_P(dummy));
+
 		convert_to_string_ex(dummy);
 		xdebug_update_ide_key(Z_STRVAL_P(dummy));
 
-		xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, Z_STRVAL_P(dummy), Z_STRLEN_P(dummy), time(NULL) + XDEBUG_COOKIE_EXPIRE_TIME, "/", 1, NULL, 0, 0, 1, 0);
+		xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, Z_STRVAL_P(dummy), Z_STRLEN_P(dummy), 0, "/", 1, NULL, 0, 0, 1, 0);
 		activate_session = 1;
 	} else if (
 		(dummy_env = getenv("XDEBUG_SESSION_START")) != NULL
 	) {
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "Found 'XDEBUG_SESSION_START' ENV variable, with value '%s'", dummy_env);
+
 		xdebug_update_ide_key(dummy_env);
 
 		if (!SG(headers_sent)) {
-			xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, XG_DBG(ide_key), strlen(XG_DBG(ide_key)), time(NULL) + XDEBUG_COOKIE_EXPIRE_TIME, "/", 1, NULL, 0, 0, 1, 0);
+			xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, XG_DBG(ide_key), strlen(XG_DBG(ide_key)), 0, "/", 1, NULL, 0, 0, 1, 0);
 		}
 
-		activate_session = 1;
-	} else if (
-		(dummy = zend_hash_str_find(Z_ARR(PG(http_globals)[TRACK_VARS_COOKIE]), "XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1)) != NULL
-	) {
-		convert_to_string_ex(dummy);
-		xdebug_update_ide_key(Z_STRVAL_P(dummy));
-
 		activate_session = 1;
 	} else if (getenv("XDEBUG_CONFIG")) {
+		xdebug_log(XLOG_CHAN_DEBUG, XLOG_DEBUG, "Found 'XDEBUG_CONFIG' ENV variable");
+
 		if (XG_DBG(ide_key) && *XG_DBG(ide_key) && !SG(headers_sent)) {
-			xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, XG_DBG(ide_key), strlen(XG_DBG(ide_key)), time(NULL) + XDEBUG_COOKIE_EXPIRE_TIME, "/", 1, NULL, 0, 0, 1, 0);
+			xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, XG_DBG(ide_key), strlen(XG_DBG(ide_key)), 0, "/", 1, NULL, 0, 0, 1, 0);
 			activate_session = 1;
 		}
 	}
@@ -626,7 +652,7 @@ static void xdebug_handle_stop_session()
 		))
 		&& !SG(headers_sent)
 	) {
-		xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, (char*) "", 0, time(NULL) + XDEBUG_COOKIE_EXPIRE_TIME, "/", 1, NULL, 0, 0, 1, 0);
+		xdebug_setcookie("XDEBUG_SESSION", sizeof("XDEBUG_SESSION") - 1, (char*) "", 0, 0, "/", 1, NULL, 0, 0, 1, 0);
 	}
 }
 
