@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,9 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   |          Nikita Popov <nikita.ppv@gmail.com>                         |
-   +----------------------------------------------------------------------+
  */
 
 #include "php.h"
@@ -273,42 +270,32 @@ static char* prepare_search_key(char *name, unsigned int *name_length, const cha
 	return element;
 }
 
-#if PHP_VERSION_ID >= 70400
-static zval *get_arrayobject_storage(zval *parent, HashTable **properties)
+char *replace_star_by_null(const char *name, int name_length)
 {
-	*properties = zend_get_properties_for(parent, ZEND_PROP_PURPOSE_DEBUG);
-	return zend_hash_str_find(*properties, "\0ArrayObject\0storage", sizeof("*ArrayObject*storage") - 1);
+	char *tmp = xdstrdup(name);
+	int   i;
+
+	for (i = 0; i < name_length; i++) {
+		if (tmp[i] == '*') {
+			tmp[i] = '\0';
+		}
+	}
+
+	return tmp;
 }
 
+#if PHP_VERSION_ID >= 70400
 static zval *get_splobjectstorage_storage(zval *parent, HashTable **properties)
 {
 	*properties = zend_get_properties_for(parent, ZEND_PROP_PURPOSE_DEBUG);
 	return zend_hash_str_find(*properties, "\0SplObjectStorage\0storage", sizeof("*SplObjectStorage*storage") - 1);
 }
-
-static zval *get_arrayiterator_storage(zval *parent, HashTable **properties)
-{
-	*properties = zend_get_properties_for(parent, ZEND_PROP_PURPOSE_DEBUG);
-	return zend_hash_str_find(*properties, "\0ArrayIterator\0storage", sizeof("*ArrayIterator*storage") - 1);
-}
 #else
-static zval *get_arrayobject_storage(zval *parent, HashTable **properties, int *is_temp)
-{
-	*properties = Z_OBJDEBUG_P(parent, *is_temp);
-	return zend_hash_str_find(*properties, "\0ArrayObject\0storage", sizeof("*ArrayObject*storage") - 1);
-}
-
 static zval *get_splobjectstorage_storage(zval *parent, HashTable **properties, int *is_temp)
 {
 	*properties = Z_OBJDEBUG_P(parent, *is_temp);
 	return zend_hash_str_find(*properties, "\0SplObjectStorage\0storage", sizeof("*SplObjectStorage*storage") - 1);
 }
-
-static zval *get_arrayiterator_storage(zval *parent, HashTable **properties, int *is_temp)
-{
-	*properties = Z_OBJDEBUG_P(parent, *is_temp);
-	return zend_hash_str_find(*properties, "\0ArrayIterator\0storage", sizeof("*ArrayIterator*storage") - 1);
-}
 #endif
 
 #if PHP_VERSION_ID < 70400
@@ -356,7 +343,7 @@ static void fetch_zval_from_symbol_table(
 			}
 #endif
 			element = prepare_search_key(name, &element_length, "", 0);
-			if (cce && ((zpp = zend_hash_str_find_ptr(&cce->properties_info, element, element_length)) != NULL) && CE_STATIC_MEMBERS(cce)) {
+			if (cce && ((zpp = zend_hash_str_find_ptr(&cce->properties_info, element, element_length)) != NULL) && cce->default_static_members_count && CE_STATIC_MEMBERS(cce)) {
 				ZVAL_COPY(&tmp_retval, &CE_STATIC_MEMBERS(cce)[zpp->offset]);
 				goto cleanup;
 			}
@@ -459,6 +446,7 @@ static void fetch_zval_from_symbol_table(
 				myht = xdebug_objdebug_pp(&value_in, &is_temp);
 #endif
 				if (myht) {
+					/* As a normal (public) property */
 					zval *tmp = zend_symtable_str_find(myht, name, name_length);
 					if (tmp != NULL) {
 						ZVAL_COPY(&tmp_retval, tmp);
@@ -469,6 +457,24 @@ static void fetch_zval_from_symbol_table(
 #endif
 						goto cleanup;
 					}
+
+					/* As a private property */
+					{
+						char *unmangled = replace_star_by_null(name, name_length);
+						zval *tmp = zend_symtable_str_find(myht, unmangled, name_length);
+						if (tmp != NULL) {
+							ZVAL_COPY(&tmp_retval, tmp);
+#if PHP_VERSION_ID >= 70400
+							zend_release_properties(myht);
+#else
+							xdebug_var_maybe_destroy_ht(myht, is_temp);
+#endif
+							xdfree(unmangled);
+							goto cleanup;
+						}
+						xdfree(unmangled);
+					}
+
 #if PHP_VERSION_ID >= 70400
 					zend_release_properties(myht);
 #else
@@ -553,75 +559,6 @@ static void fetch_zval_from_symbol_table(
 #endif
 			}
 
-			/* Then we try to see whether the first char is * and use the part between * and * as class name for the private property */
-			if (name[0] == '*') {
-				char *secondStar;
-
-				secondStar = strstr(name + 1, "*");
-				if (secondStar) {
-					free(element);
-					element_length = name_length - (secondStar + 1 - name);
-
-					/* All right, time for a mega hack. It's ArrayObject access time! */
-					if (strncmp(name + 1, "ArrayObject", secondStar - name - 1) == 0 && strncmp(secondStar + 1, "storage", element_length) == 0) {
-#if PHP_VERSION_ID >= 70400
-						zval *tmp = get_arrayobject_storage(value_in, &myht);
-#else
-						zval *tmp = get_arrayobject_storage(value_in, &myht, &is_temp);
-#endif
-						element = NULL;
-						if (tmp != NULL) {
-							ZVAL_COPY(&tmp_retval, tmp);
-#if PHP_VERSION_ID >= 70400
-							zend_release_properties(myht);
-#else
-							xdebug_var_maybe_destroy_ht(myht, is_temp);
-#endif
-							goto cleanup;
-						}
-#if PHP_VERSION_ID >= 70400
-						zend_release_properties(myht);
-#else
-						xdebug_var_maybe_destroy_ht(myht, is_temp);
-#endif
-					}
-					/* All right, time for a mega hack. It's ArrayIterator access time! */
-					if (strncmp(name + 1, "ArrayIterator", secondStar - name - 1) == 0 && strncmp(secondStar + 1, "storage", element_length) == 0) {
-#if PHP_VERSION_ID >= 70400
-						zval *tmp = get_arrayiterator_storage(value_in, &myht);
-#else
-						zval *tmp = get_arrayiterator_storage(value_in, &myht, &is_temp);
-#endif
-						element = NULL;
-						if (tmp != NULL) {
-							ZVAL_COPY(&tmp_retval, tmp);
-#if PHP_VERSION_ID >= 70400
-							zend_release_properties(myht);
-#else
-							xdebug_var_maybe_destroy_ht(myht, is_temp);
-#endif
-							goto cleanup;
-						}
-#if PHP_VERSION_ID >= 70400
-						zend_release_properties(myht);
-#else
-						xdebug_var_maybe_destroy_ht(myht, is_temp);
-#endif
-
-					}
-
-					/* The normal one */
-					element = prepare_search_key(secondStar + 1, &element_length, name + 1, secondStar - name - 1);
-					if (ht) {
-						zval *tmp = zend_hash_str_find(ht, element, element_length);
-						if (tmp != NULL) {
-							ZVAL_COPY(&tmp_retval, tmp);
-							goto cleanup;
-						}
-					}
-				}
-			}
-
 			break;
 	}
 
@@ -836,8 +773,8 @@ void xdebug_get_php_symbol(zval *retval, xdebug_str* name)
 						state = 1;
 						keyword_end = &ptr[ctr];
 
-						if (strncmp(keyword, "::", 2) == 0 && active_fse->function.class_name) { /* static class properties */
-							zend_class_entry *ce = zend_fetch_class(active_fse->function.class_name, ZEND_FETCH_CLASS_SELF);
+						if (strncmp(keyword, "::", 2) == 0 && active_fse->function.object_class) { /* static class properties */
+							zend_class_entry *ce = zend_fetch_class(active_fse->function.object_class, ZEND_FETCH_CLASS_SELF);
 
 							current_classname = estrdup(STR_NAME_VAL(ce->name));
 							cc_length = strlen(STR_NAME_VAL(ce->name));
@@ -916,8 +853,18 @@ xdebug_str* xdebug_get_property_type(zval* object, zval *val)
 	if (info) {
 #if PHP_VERSION_ID >= 80000
 		zend_string *type_info = zend_type_to_string(info->type);
-		type_str = xdebug_str_create(ZSTR_VAL(type_info), ZSTR_LEN(type_info));
+		type_str = xdebug_str_new();
+
+# if PHP_VERSION_ID >= 80100
+		if (info->flags & ZEND_ACC_READONLY) {
+			xdebug_str_add_literal(type_str, "readonly ");
+		}
+# endif
+
+		xdebug_str_add_zstr(type_str, type_info);
+
 		zend_string_release(type_info);
+
 #else
 		type_str = xdebug_str_new();
 
@@ -1142,9 +1089,9 @@ static char* xdebug_create_doc_link(xdebug_func f)
 		case XFUNC_STATIC_MEMBER:
 		case XFUNC_MEMBER: {
 			if (strcmp(f.function, "__construct") == 0) {
-				tmp_target = xdebug_sprintf("%s.construct", ZSTR_VAL(f.class_name));
+				tmp_target = xdebug_sprintf("%s.construct", ZSTR_VAL(f.object_class));
 			} else {
-				tmp_target = xdebug_sprintf("%s.%s", ZSTR_VAL(f.class_name), f.function);
+				tmp_target = xdebug_sprintf("%s.%s", ZSTR_VAL(f.object_class), f.function);
 			}
 			break;
 		}
@@ -1154,7 +1101,7 @@ static char* xdebug_create_doc_link(xdebug_func f)
 		*p = '-';
 	}
 
-	retval = xdebug_sprintf("<a href='%s%s%s' target='_new'>%s</a>\n",
+	retval = xdebug_sprintf("<a href='%s%s%s' target='_new'>%s</a>",
 		(PG(docref_root) && PG(docref_root)[0]) ? PG(docref_root) : "http://www.php.net/",
 		tmp_target, PG(docref_ext), f.function);
 
@@ -1163,11 +1110,11 @@ static char* xdebug_create_doc_link(xdebug_func f)
 	return retval;
 }
 
-char* xdebug_show_fname(xdebug_func f, int html, int flags)
+char* xdebug_show_fname(xdebug_func f, int flags)
 {
 	switch (f.type) {
 		case XFUNC_NORMAL: {
-			if (PG(html_errors) && html && f.internal) {
+			if (PG(html_errors) && (flags & XDEBUG_SHOW_FNAME_ALLOW_HTML) && f.internal) {
 				return xdebug_create_doc_link(f);
 			} else {
 				return xdstrdup(f.function);
@@ -1177,11 +1124,19 @@ char* xdebug_show_fname(xdebug_func f, int html, int flags)
 
 		case XFUNC_STATIC_MEMBER:
 		case XFUNC_MEMBER: {
-			if (PG(html_errors) && html && f.internal) {
+			if (PG(html_errors) && (flags & XDEBUG_SHOW_FNAME_ALLOW_HTML) && f.internal) {
 				return xdebug_create_doc_link(f);
 			} else {
+				if (f.scope_class && !(flags & XDEBUG_SHOW_FNAME_IGNORE_SCOPE)) {
+					return xdebug_sprintf("%s%s%s",
+						ZSTR_VAL(f.scope_class),
+						f.type == XFUNC_STATIC_MEMBER ? "::" : "->",
+						f.function ? f.function : "?"
+					);
+				}
+
 				return xdebug_sprintf("%s%s%s",
-					f.class_name ? ZSTR_VAL(f.class_name) : "?",
+					f.object_class ? ZSTR_VAL(f.object_class) : "?",
 					f.type == XFUNC_STATIC_MEMBER ? "::" : "->",
 					f.function ? f.function : "?"
 				);
@@ -1217,6 +1172,12 @@ char* xdebug_show_fname(xdebug_func f, int html, int flags)
 			return xdstrdup("{zend_pass}");
 			break;
 
+#if PHP_VERSION_ID >= 80100
+		case XFUNC_FIBER:
+			return xdstrdup(f.function);
+			break;
+#endif
+
 		default:
 			return xdstrdup("{unknown}");
 	}
