@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,12 +12,14 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   +----------------------------------------------------------------------+
  */
 
 #include "php.h"
 #include "ext/standard/php_string.h"
+#include "Zend/zend_closures.h"
+#if PHP_VERSION_ID >= 80100
+# include "zend_enum.h"
+#endif
 
 #include "var_export_line.h"
 
@@ -31,28 +33,27 @@ static int xdebug_array_element_export(zval *zv_nptr, zend_ulong index_key, zend
 		options->runtime[level].current_element_nr < options->runtime[level].end_element_nr)
 	{
 		if (HASH_KEY_IS_NUMERIC(hash_key)) { /* numeric key */
-			xdebug_str_add(str, xdebug_sprintf(XDEBUG_INT_FMT " => ", index_key), 1);
+			xdebug_str_add_fmt(str, XDEBUG_INT_FMT " => ", index_key);
 		} else { /* string key */
-			size_t newlen = 0;
-			char *tmp, *tmp2;
+			zend_string *tmp, *tmp2;
 
-			tmp = xdebug_str_to_str((char*) HASH_APPLY_KEY_VAL(hash_key), HASH_APPLY_KEY_LEN(hash_key), "'", 1, "\\'", 2, &newlen);
-			tmp2 = xdebug_str_to_str(tmp, newlen - 1, "\0", 1, "\\0", 2, &newlen);
+			tmp = php_str_to_str(ZSTR_VAL(hash_key), ZSTR_LEN(hash_key), (char*) "'", 1, (char*) "\\'", 2);
+			tmp2 = php_str_to_str(ZSTR_VAL(tmp), ZSTR_LEN(tmp), (char*) "\0", 1, (char*) "\\0", 2);
 			if (tmp) {
-				efree(tmp);
+				zend_string_release(tmp);
 			}
-			xdebug_str_addl(str, "'", 1, 0);
+			xdebug_str_addc(str, '\'');
 			if (tmp2) {
-				xdebug_str_addl(str, tmp2, newlen, 0);
-				efree(tmp2);
+				xdebug_str_add_zstr(str, tmp2);
+				zend_string_release(tmp2);
 			}
-			xdebug_str_add(str, "' => ", 0);
+			xdebug_str_add_literal(str, "' => ");
 		}
 		xdebug_var_export_line(zv, str, level + 2, debug_zval, options);
-		xdebug_str_addl(str, ", ", 2, 0);
+		xdebug_str_add_literal(str, ", ");
 	}
 	if (options->runtime[level].current_element_nr == options->runtime[level].end_element_nr) {
-		xdebug_str_addl(str, "..., ", 5, 0);
+		xdebug_str_add_literal(str, "..., ");
 	}
 	options->runtime[level].current_element_nr++;
 	return 0;
@@ -80,16 +81,16 @@ static int xdebug_object_element_export(zval *object, zval *zv_nptr, zend_ulong
 				xdebug_str_addc(str, ' ');
 				xdebug_str_add_str(str, property_type);
 			}
-			xdebug_str_addl(str, " $", 2, 0);
+			xdebug_str_add_literal(str, " $");
 			if (strcmp(modifier, "private") != 0 || strcmp(class_name, prop_class_name) == 0) {
 				xdebug_str_add_str(str, property_name);
-				xdebug_str_addl(str, " = ", 3, 0);
+				xdebug_str_add_literal(str, " = ");
 			} else {
 				xdebug_str_addc(str, '{');
 				xdebug_str_add(str, prop_class_name, 0);
 				xdebug_str_addc(str, '}');
 				xdebug_str_add_str(str, property_name);
-				xdebug_str_addl(str, " = ", 3, 0);
+				xdebug_str_add_literal(str, " = ");
 			}
 
 			if (property_type) {
@@ -98,22 +99,50 @@ static int xdebug_object_element_export(zval *object, zval *zv_nptr, zend_ulong
 			xdebug_str_free(property_name);
 			xdfree(prop_class_name);
 		} else {
-			xdebug_str_add(str, xdebug_sprintf("public $%d = ", index_key), 1);
+			xdebug_str_add_fmt(str, "public $%d = ", index_key);
 		}
 		xdebug_var_export_line(zv, str, level + 2, debug_zval, options);
-		xdebug_str_addl(str, "; ", 2, 0);
+		xdebug_str_add_literal(str, "; ");
 	}
 	if (options->runtime[level].current_element_nr == options->runtime[level].end_element_nr) {
-		xdebug_str_addl(str, "...; ", 5, 0);
+		xdebug_str_add_literal(str, "...; ");
 	}
 	options->runtime[level].current_element_nr++;
 	return 0;
 }
 
+static void handle_closure(xdebug_str *str, zval *obj)
+{
+	const zend_function *closure_function;
+
+	if (Z_TYPE_P(obj) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(obj), zend_ce_closure)) {
+		return;
+	}
+
+#if PHP_VERSION_ID >= 80000
+	closure_function = zend_get_closure_method_def(Z_OBJ_P(obj));
+#else
+	closure_function = zend_get_closure_method_def(obj);
+#endif
+
+	xdebug_str_add_literal(str, "virtual $closure = \"");
+
+	if (closure_function->common.scope) {
+		if (closure_function->common.fn_flags & ZEND_ACC_STATIC) {
+			xdebug_str_add_zstr(str, closure_function->common.scope->name);
+			xdebug_str_add_literal(str, "::");
+		} else {
+			xdebug_str_add_literal(str, "$this->");
+		}
+	}
+	xdebug_str_add_zstr(str, closure_function->common.function_name);
+
+	xdebug_str_add_literal(str, "\", ");
+}
+
 void xdebug_var_export_line(zval **struc, xdebug_str *str, int level, int debug_zval, xdebug_var_export_options *options)
 {
 	HashTable *myht;
-	char*     tmp_str;
 #if PHP_VERSION_ID < 70400
 	int       is_temp;
 #endif
@@ -121,72 +150,81 @@ void xdebug_var_export_line(zval **struc, xdebug_str *str, int level, int debug_
 	zend_string *key;
 	zval *val;
 	zval *tmpz;
+	int   z_type;
 
 	if (!struc || !(*struc)) {
 		return;
 	}
 
+	z_type = Z_TYPE_P(*struc);
+
 	if (debug_zval) {
 		xdebug_add_variable_attributes(str, *struc, XDEBUG_VAR_ATTR_TEXT);
 	}
-	if (Z_TYPE_P(*struc) == IS_INDIRECT) {
+	if (z_type == IS_INDIRECT) {
 		tmpz = Z_INDIRECT_P(*struc);
 		struc = &tmpz;
+		z_type = Z_TYPE_P(*struc);
 	}
-	if (Z_TYPE_P(*struc) == IS_REFERENCE) {
+	if (z_type == IS_REFERENCE) {
 		tmpz = &((*struc)->value.ref->val);
 		struc = &tmpz;
+		z_type = Z_TYPE_P(*struc);
 	}
 
-	switch (Z_TYPE_P(*struc)) {
+	switch (z_type) {
 		case IS_TRUE:
+			xdebug_str_add_literal(str, "TRUE");
+			break;
+
 		case IS_FALSE:
-			xdebug_str_add(str, xdebug_sprintf("%s", Z_TYPE_P(*struc) == IS_TRUE ? "TRUE" : "FALSE"), 1);
+			xdebug_str_add_literal(str, "FALSE");
 			break;
 
 		case IS_NULL:
-			xdebug_str_addl(str, "NULL", 4, 0);
+			xdebug_str_add_literal(str, "NULL");
 			break;
 
 		case IS_LONG:
-			xdebug_str_add(str, xdebug_sprintf(XDEBUG_INT_FMT, Z_LVAL_P(*struc)), 1);
+			xdebug_str_add_fmt(str, XDEBUG_INT_FMT, Z_LVAL_P(*struc));
 			break;
 
 		case IS_DOUBLE:
-			xdebug_str_add(str, xdebug_sprintf("%.*G", (int) EG(precision), Z_DVAL_P(*struc)), 1);
+			xdebug_str_add_fmt(str, "%.*G", (int) EG(precision), Z_DVAL_P(*struc));
 			break;
 
 		case IS_STRING: {
-			zend_string *i_string = zend_string_init(Z_STRVAL_P(*struc), Z_STRLEN_P(*struc), 0);
 			zend_string *tmp_zstr;
 
 #if PHP_VERSION_ID >= 70300
-			tmp_zstr = php_addcslashes(i_string, (char*) "'\\\0..\37", 7);
+			tmp_zstr = php_addcslashes(Z_STR_P(*struc), (char*) "'\\\0..\37", 7);
 #else
-			tmp_zstr = php_addcslashes(i_string, 0, (char*) "'\\\0..\37", 7);
+			tmp_zstr = php_addcslashes(Z_STR_P(*struc), 0, (char*) "'\\\0..\37", 7);
 #endif
 
-			tmp_str = estrndup(tmp_zstr->val, tmp_zstr->len);
-			zend_string_release(tmp_zstr);
-			zend_string_release(i_string);
-
 			if (options->no_decoration) {
-				xdebug_str_add(str, tmp_str, 0);
+				xdebug_str_add_zstr(str, tmp_zstr);
 			} else if ((size_t) Z_STRLEN_P(*struc) <= (size_t) options->max_data) {
-				xdebug_str_add(str, xdebug_sprintf("'%s'", tmp_str), 1);
+				xdebug_str_addc(str, '\'');
+				xdebug_str_add_zstr(str, tmp_zstr);
+				xdebug_str_addc(str, '\'');
 			} else {
-				xdebug_str_addl(str, "'", 1, 0);
-				xdebug_str_addl(str, xdebug_sprintf("%s", tmp_str), options->max_data, 1);
-				xdebug_str_addl(str, "...'", 4, 0);
+				xdebug_str_addc(str, '\'');
+				xdebug_str_addl(str, ZSTR_VAL(tmp_zstr), options->max_data, 0);
+				xdebug_str_addc(str, '\'');
 			}
-			efree(tmp_str);
+			zend_string_release(tmp_zstr);
 		} break;
 
 		case IS_ARRAY:
 			myht = Z_ARRVAL_P(*struc);
 
 			if (!xdebug_zend_hash_is_recursive(myht)) {
-				xdebug_str_addl(str, "array (", 7, 0);
+				if (debug_zval) {
+					xdebug_str_add_literal(str, "array (");
+				} else {
+					xdebug_str_addc(str, '[');
+				}
 				if (level <= options->max_depth) {
 					options->runtime[level].current_element_nr = 0;
 					options->runtime[level].start_element_nr = 0;
@@ -205,26 +243,57 @@ void xdebug_var_export_line(zval **struc, xdebug_str *str, int level, int debug_
 						xdebug_str_chop(str, 2);
 					}
 				} else {
-					xdebug_str_addl(str, "...", 3, 0);
+					xdebug_str_add_literal(str, "...");
 				}
-				xdebug_str_addl(str, ")", 1, 0);
+				xdebug_str_addc(str, debug_zval ? ')' : ']');
 			} else {
-				xdebug_str_addl(str, "...", 3, 0);
+				xdebug_str_add_literal(str, "...");
 			}
 			break;
 
-		case IS_OBJECT:
+		case IS_OBJECT: {
+#if PHP_VERSION_ID >= 80100
+			zend_class_entry *ce = Z_OBJCE_P(*struc);
+			if (ce->ce_flags & ZEND_ACC_ENUM) {
+				zval *case_name_zval = zend_enum_fetch_case_name(Z_OBJ_P(*struc));
+				xdebug_str_add_fmt(str, "enum %s::%s", ZSTR_VAL(ce->name), Z_STRVAL_P(case_name_zval));
+
+				if (ce->enum_backing_type != IS_UNDEF) {
+					zval *value = zend_enum_fetch_case_value(Z_OBJ_P(*struc));
+
+					if (ce->enum_backing_type == IS_LONG) {
+						xdebug_str_add_fmt(str, "(" XDEBUG_INT_FMT ")", Z_LVAL_P(value));
+					}
+
+					if (ce->enum_backing_type == IS_STRING) {
+						zend_string *tmp_zstr = php_addcslashes(Z_STR_P(value), (char*) "'\\\0..\37", 7);
+
+						xdebug_str_add_literal(str, "('");
+						xdebug_str_add_zstr(str, tmp_zstr);
+						xdebug_str_add_literal(str, "')");
+
+						zend_string_release(tmp_zstr);
+					}
+				}
+
+				break;
+			}
+#endif
+
 #if PHP_VERSION_ID >= 70400
 			myht = xdebug_objdebug_pp(struc);
 #else
 			myht = xdebug_objdebug_pp(struc, &is_temp);
 #endif
 
-			if (!xdebug_zend_hash_is_recursive(myht)) {
-				char *class_name = (char*) STR_NAME_VAL(Z_OBJCE_P(*struc)->name);
-				xdebug_str_add(str, xdebug_sprintf("class %s { ", class_name), 1);
+			if (!myht || !xdebug_zend_hash_is_recursive(myht)) {
+				xdebug_str_add_literal(str, "class ");
+				xdebug_str_add(str, ZSTR_VAL(Z_OBJCE_P(*struc)->name), 0);
+				xdebug_str_add_literal(str, " { ");
 
-				if (level <= options->max_depth) {
+				handle_closure(str, *struc);
+
+				if (myht && (level <= options->max_depth)) {
 					options->runtime[level].current_element_nr = 0;
 					options->runtime[level].start_element_nr = 0;
 					options->runtime[level].end_element_nr = options->max_children;
@@ -232,7 +301,7 @@ void xdebug_var_export_line(zval **struc, xdebug_str *str, int level, int debug_
 					xdebug_zend_hash_apply_protection_begin(myht);
 
 					ZEND_HASH_FOREACH_KEY_VAL(myht, num, key, val) {
-						xdebug_object_element_export(*struc, val, num, key, level, str, debug_zval, options, class_name);
+						xdebug_object_element_export(*struc, val, num, key, level, str, debug_zval, options, ZSTR_VAL(Z_OBJCE_P(*struc)->name));
 					} ZEND_HASH_FOREACH_END();
 
 					xdebug_zend_hash_apply_protection_end(myht);
@@ -242,11 +311,11 @@ void xdebug_var_export_line(zval **struc, xdebug_str *str, int level, int debug_
 						xdebug_str_chop(str, 2);
 					}
 				} else {
-					xdebug_str_addl(str, "...", 3, 0);
+					xdebug_str_add_literal(str, "...");
 				}
-				xdebug_str_addl(str, " }", 2, 0);
+				xdebug_str_add_literal(str, " }");
 			} else {
-				xdebug_str_addl(str, "...", 3, 0);
+				xdebug_str_add_literal(str, "...");
 			}
 #if PHP_VERSION_ID >= 70400
 			zend_release_properties(myht);
@@ -254,21 +323,22 @@ void xdebug_var_export_line(zval **struc, xdebug_str *str, int level, int debug_
 			xdebug_var_maybe_destroy_ht(myht, is_temp);
 #endif
 			break;
+		}
 
 		case IS_RESOURCE: {
 			char *type_name;
 
 			type_name = (char *) zend_rsrc_list_get_rsrc_type(Z_RES_P(*struc));
-			xdebug_str_add(str, xdebug_sprintf("resource(%ld) of type (%s)", Z_RES_P(*struc)->handle, type_name ? type_name : "Unknown"), 1);
+			xdebug_str_add_fmt(str, "resource(%ld) of type (%s)", Z_RES_P(*struc)->handle, type_name ? type_name : "Unknown");
 			break;
 		}
 
 		case IS_UNDEF:
-			xdebug_str_addl(str, "*uninitialized*", sizeof("*uninitialized*") - 1, 0);
+			xdebug_str_add_literal(str, "*uninitialized*");
 			break;
 
 		default:
-			xdebug_str_addl(str, "NFC", 3, 0);
+			xdebug_str_add_literal(str, "NFC");
 			break;
 	}
 }
@@ -312,36 +382,37 @@ static void xdebug_var_synopsis(zval **struc, xdebug_str *str, int level, int de
 
 	switch (Z_TYPE_P(*struc)) {
 		case IS_TRUE:
-			xdebug_str_addl(str, "true", 4, 0);
+			xdebug_str_add_literal(str, "true");
 			break;
 
 		case IS_FALSE:
-			xdebug_str_addl(str, "false", 5, 0);
+			xdebug_str_add_literal(str, "false");
 			break;
 
 		case IS_NULL:
-			xdebug_str_addl(str, "null", 4, 0);
+			xdebug_str_add_literal(str, "null");
 			break;
 
 		case IS_LONG:
-			xdebug_str_addl(str, "long", 4, 0);
+			xdebug_str_add_literal(str, "long");
 			break;
 
 		case IS_DOUBLE:
-			xdebug_str_addl(str, "double", 6, 0);
+			xdebug_str_add_literal(str, "double");
 			break;
 
 		case IS_STRING:
-			xdebug_str_add(str, xdebug_sprintf("string(%d)", Z_STRLEN_P(*struc)), 1);
+			xdebug_str_add_fmt(str, "string(%d)", Z_STRLEN_P(*struc));
 			break;
 
 		case IS_ARRAY:
 			myht = Z_ARRVAL_P(*struc);
-			xdebug_str_add(str, xdebug_sprintf("array(%d)", myht->nNumOfElements), 1);
+			xdebug_str_add_fmt(str, "array(%d)", myht->nNumOfElements);
 			break;
 
 		case IS_OBJECT: {
-			xdebug_str_add(str, xdebug_sprintf("class %s", STR_NAME_VAL(Z_OBJCE_P(*struc)->name)), 1);
+			xdebug_str_add_literal(str, "class ");
+			xdebug_str_add(str, ZSTR_VAL(Z_OBJCE_P(*struc)->name), 0);
 			break;
 		}
 
@@ -349,16 +420,16 @@ static void xdebug_var_synopsis(zval **struc, xdebug_str *str, int level, int de
 			char *type_name;
 
 			type_name = (char *) zend_rsrc_list_get_rsrc_type(Z_RES_P(*struc));
-			xdebug_str_add(str, xdebug_sprintf("resource(%ld) of type (%s)", Z_RES_P(*struc)->handle, type_name ? type_name : "Unknown"), 1);
+			xdebug_str_add_fmt(str, "resource(%ld) of type (%s)", Z_RES_P(*struc)->handle, type_name ? type_name : "Unknown");
 			break;
 		}
 
 		case IS_UNDEF:
-			xdebug_str_addl(str, "*uninitialized*", sizeof("*uninitialized*") - 1, 0);
+			xdebug_str_add_literal(str, "*uninitialized*");
 			break;
 
 		default:
-			xdebug_str_addl(str, "NFC", 3, 0);
+			xdebug_str_add_literal(str, "NFC");
 			break;
 
 	}
