@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2020 Derick Rethans                               |
+   | Copyright (c) 2002-2021 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -12,9 +12,6 @@
    | to obtain it through the world-wide-web, please send a note to       |
    | derick@xdebug.org so we can mail you a copy immediately.             |
    +----------------------------------------------------------------------+
-   | Authors: Derick Rethans <derick@xdebug.org>                          |
-   |          Michael Voříšek <mvorisek@mvorisek.cz>                      |
-   +----------------------------------------------------------------------+
  */
 
 #ifdef PHP_WIN32
@@ -57,8 +54,7 @@ void xdebug_profiler_mshutdown(void)
 
 void xdebug_profiler_rinit(void)
 {
-	XG_PROF(profile_file)  = NULL;
-	XG_PROF(profile_filename) = NULL;
+	xdebug_file_init(&XG_PROF(profile_file));
 	XG_PROF(profile_filename_refs) = NULL;
 	XG_PROF(profile_functionname_refs) = NULL;
 	XG_PROF(profile_last_filename_ref) = 0;
@@ -161,15 +157,15 @@ void xdebug_profile_call_entry_dtor(void *dummy, void *elem)
 	xdfree(ce);
 }
 
-static void profiler_write_header(FILE *file, char *script_name)
+static void profiler_write_header(xdebug_file *file, char *script_name)
 {
 	if (XINI_PROF(profiler_append)) {
-		fprintf(file, "\n==== NEW PROFILING FILE ==============================================\n");
+		xdebug_file_printf(file, "\n==== NEW PROFILING FILE ==============================================\n");
 	}
-	fprintf(file, "version: 1\ncreator: xdebug %s (PHP %s)\n", XDEBUG_VERSION, PHP_VERSION);
-	fprintf(file, "cmd: %s\npart: 1\npositions: line\n\n", script_name);
-	fprintf(file, "events: Time_(10ns) Memory_(bytes)\n\n");
-	fflush(file);
+	xdebug_file_printf(file, "version: 1\ncreator: xdebug %s (PHP %s)\n", XDEBUG_VERSION, XG_BASE(php_version_run_time));
+	xdebug_file_printf(file, "cmd: %s\npart: 1\npositions: line\n\n", script_name);
+	xdebug_file_printf(file, "events: Time_(10ns) Memory_(bytes)\n\n");
+	xdebug_file_flush(file);
 }
 
 #define NANOTIME_SCALE_10NS(nanotime) ((unsigned long)(((nanotime) + 5) / 10))
@@ -199,23 +195,17 @@ void xdebug_profiler_init(char *script_name)
 		filename = xdebug_sprintf("%s%c%s", output_dir, DEFAULT_SLASH, fname);
 	}
 
-	if (XINI_PROF(profiler_append)) {
-		XG_PROF(profile_file) = xdebug_fopen(filename, "a", NULL, &XG_PROF(profile_filename));
-	} else {
-		XG_PROF(profile_file) = xdebug_fopen(filename, "w", NULL, &XG_PROF(profile_filename));
-	}
-
-	if (!XG_PROF(profile_file)) {
-		xdebug_log_diagnose_permissions(XLOG_CHAN_PROFILE, output_dir, fname);
+	if (!xdebug_file_open(&XG_PROF(profile_file), filename, NULL, XINI_PROF(profiler_append) ? "a" : "w")) {
+		xdebug_log_diagnose_permissions(XLOG_CHAN_PROFILE, output_dir, XG_PROF(profile_file).name);
 		goto return_and_free_names;
 	}
 
-	profiler_write_header(XG_PROF(profile_file), script_name);
+	profiler_write_header(&XG_PROF(profile_file), script_name);
 
 	if (!SG(headers_sent)) {
 		sapi_header_line ctr = {0};
 
-		ctr.line = xdebug_sprintf("X-Xdebug-Profile-Filename: %s", XG_PROF(profile_filename));
+		ctr.line = xdebug_sprintf("X-Xdebug-Profile-Filename: %s", XG_PROF(profile_file).name);
 		ctr.line_len = strlen(ctr.line);
 		sapi_header_op(SAPI_HEADER_REPLACE, &ctr);
 		xdfree((void*) ctr.line);
@@ -243,8 +233,8 @@ void xdebug_profiler_deinit()
 		xdebug_profiler_function_end(fse);
 	}
 
-	fprintf(
-		XG_PROF(profile_file),
+	xdebug_file_printf(
+		&XG_PROF(profile_file),
 		"summary: %lu %zd\n\n",
 		NANOTIME_SCALE_10NS(xdebug_get_nanotime() - XG_PROF(profiler_start_nanotime)),
 		zend_memory_peak_usage(0)
@@ -252,15 +242,11 @@ void xdebug_profiler_deinit()
 
 	XG_PROF(active) = 0;
 
-	fflush(XG_PROF(profile_file));
+	xdebug_file_flush(&XG_PROF(profile_file));
 
-	if (XG_PROF(profile_file)) {
-		fclose(XG_PROF(profile_file));
-		XG_PROF(profile_file) = NULL;
-	}
-
-	if (XG_PROF(profile_filename)) {
-		xdfree(XG_PROF(profile_filename));
+	if (XG_PROF(profile_file).type != XDEBUG_FILE_TYPE_NULL) {
+		xdebug_file_close(&XG_PROF(profile_file));
+		xdebug_file_deinit(&XG_PROF(profile_file));
 	}
 
 	xdebug_hash_destroy(XG_PROF(profile_filename_refs));
@@ -327,7 +313,7 @@ void xdebug_profiler_add_function_details_user(function_stack_entry *fse, zend_o
 {
 	char *tmp_fname, *tmp_name;
 
-	tmp_name = xdebug_show_fname(fse->function, 0, 0);
+	tmp_name = xdebug_show_fname(fse->function, XDEBUG_SHOW_FNAME_DEFAULT);
 	switch (fse->function.type) {
 		case XFUNC_INCLUDE:
 		case XFUNC_INCLUDE_ONCE:
@@ -364,7 +350,7 @@ void xdebug_profiler_add_function_details_internal(function_stack_entry *fse)
 {
 	char *tmp_fname, *tmp_name;
 
-	tmp_name = xdebug_show_fname(fse->function, 0, 0);
+	tmp_name = xdebug_show_fname(fse->function, XDEBUG_SHOW_FNAME_DEFAULT);
 	switch (fse->function.type) {
 		case XFUNC_INCLUDE:
 		case XFUNC_INCLUDE_ONCE:
@@ -529,7 +515,7 @@ void xdebug_profiler_function_end(function_stack_entry *fse)
 	}
 	xdebug_str_addc(&file_buffer, '\n');
 
-	fwrite(file_buffer.d, sizeof(char), file_buffer.l, XG_PROF(profile_file));
+	xdebug_file_write(file_buffer.d, sizeof(char), file_buffer.l, &XG_PROF(profile_file));
 	xdebug_str_dtor(file_buffer);
 }
 
@@ -553,7 +539,7 @@ char *xdebug_get_profiler_filename()
 		return NULL;
 	}
 
-	return XG_PROF(profile_filename);
+	return XG_PROF(profile_file).name;
 }
 
 PHP_FUNCTION(xdebug_get_profiler_filename)
